# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from unittest import TestCase, skipUnless
from commonconf import override_settings
from restclients_core.dao import DAO
from restclients_core.exceptions import DataFailureException
from urllib3.connectionpool import HTTPConnectionPool
from urllib3.exceptions import MaxRetryError, SSLError
import mock
import os


class TDAO(DAO):
    def service_name(self):
        return 'live_test'

    def get_default_service_setting(self, key):
        if 'DAO_CLASS' == key:
            return 'Live'

        if 'HOST' == key:
            return 'http://localhost:9876/'


class SSLTDAO(DAO):
    def service_name(self):
        return 'live_ssl_test'

    def get_default_service_setting(self, key):
        if 'DAO_CLASS' == key:
            return 'Live'

        if 'HOST' == key:
            return 'https://localhost:9443/'


class SSLClientCertTDAO(DAO):
    def service_name(self):
        return 'live_ssl_test_cert'

    def get_default_service_setting(self, key):
        if 'DAO_CLASS' == key:
            return 'Live'

        if 'HOST' == key:
            return 'https://localhost:9443/'

        if 'CERT_FILE' == key:
            return 'test/certs/client-cert.pem'

        if 'KEY_FILE' == key:
            return 'test/certs/client_key.pem'


class SSLBadFailTDAO(DAO):
    def service_name(self):
        return 'live_ssl_test_fail'

    def get_default_service_setting(self, key):
        if 'DAO_CLASS' == key:
            return 'Live'

        if 'HOST' == key:
            return 'https://localhost:9444/'


class SSLBadIgnoreTDAO(DAO):
    def service_name(self):
        return 'live_ssl_test_ignore'

    def get_default_service_setting(self, key):
        if 'DAO_CLASS' == key:
            return 'Live'

        if 'HOST' == key:
            return 'https://localhost:9444/'

        if 'VERIFY_HTTPS' == key:
            return False


@skipUnless('RUN_LIVE_TESTS' in os.environ, 'RUN_LIVE_TESTS=1 to run tests')
class TestLive(TestCase):
    def test_found_resource(self):
        response = TDAO().getURL('/ok', {})
        self.assertEqual(response.status, 200)
        self.assertEqual(response.data, b'ok')
        self.assertEqual(response.headers['X-Custom-Header'], 'header-test')
        self.assertEqual(response.headers.get('X-Custom-Header'),
                         'header-test')

    def test_clear_cached_response(self):
        self.assertIsNone(TDAO().clear_cached_response('/ok'))

    def test_missing_resource(self):
        response = TDAO().getURL('/missing.json', {})
        self.assertEqual(response.status, 404)

    def test_other_status(self):
        response = TDAO().getURL('/403', {})
        self.assertEqual(response.status, 403)

    def test_one_redirect(self):
        response = TDAO().getURL('/301', {})
        self.assertEqual(response.status, 200)
        self.assertEqual(response.data, b'ok')

    def test_multiple_redirects(self):
        self.assertRaises(DataFailureException, TDAO().getURL, '/redirect', {})

        try:
            TDAO().getURL('/redirect', {})
        except DataFailureException as ex:
            self.assertEqual(ex.url, '/redirect')
            self.assertEqual(ex.status, 0)

    @mock.patch.object(HTTPConnectionPool, 'urlopen')
    def test_max_retry_error(self, mock_urlopen):
        mock_urlopen.side_effect = MaxRetryError(None, '/ok')

        self.assertRaises(DataFailureException, TDAO().getURL, '/ok', {})

        try:
            TDAO().getURL('/ok', {})
        except DataFailureException as ex:
            self.assertEqual(ex.url, '/ok')
            self.assertEqual(ex.status, 0)

    def test_settings_defaults(self):
        live_dao = TDAO().get_implementation()

        self.assertEqual(live_dao._get_connect_timeout(), 3)

        with override_settings(RESTCLIENTS_DEFAULT_CONNECT_TIMEOUT=0.2):
            self.assertEqual(live_dao._get_connect_timeout(), 0.2)

        with override_settings(RESTCLIENTS_LIVE_TEST_CONNECT_TIMEOUT=0.1):
            self.assertEqual(live_dao._get_connect_timeout(), 0.1)

        self.assertEqual(live_dao._get_timeout(), 10)

        with override_settings(RESTCLIENTS_LIVE_TEST_TIMEOUT=5):
            self.assertEqual(live_dao._get_timeout(), 5)

        with override_settings(RESTCLIENTS_DEFAULT_TIMEOUT=0.25):
            self.assertEqual(live_dao._get_timeout(), 0.25)

        self.assertEqual(live_dao._get_max_pool_size(), 10)

        with override_settings(RESTCLIENTS_LIVE_TEST_POOL_SIZE=5):
            self.assertEqual(live_dao._get_max_pool_size(), 5)

        with override_settings(RESTCLIENTS_DEFAULT_POOL_SIZE=25):
            self.assertEqual(live_dao._get_max_pool_size(), 25)


@skipUnless('RUN_SSL_TESTS' in os.environ, 'RUN_SSL_TESTS=1 to run tests')
class TestLiveSSL(TestCase):
    def test_ssl_found_resource(self):
        response = SSLTDAO().getURL('/ok', {})
        self.assertEqual(response.status, 200)
        self.assertEqual(response.data, b'Ok')
        self.assertEqual(response.headers['X-Custom-Header'], 'header-test')
        self.assertEqual(response.headers.get('X-Custom-Header'),
                         'header-test')

    def test_ssl_client_cert(self):
        response = SSLClientCertTDAO().getURL('/ok', {})
        self.assertEqual(response.status, 200)
        self.assertEqual(response.data, b'Ok')

    def test_ssl_non_validated_cert(self):
        self.assertRaises(DataFailureException, SSLBadFailTDAO().getURL, '/')

        try:
            SSLBadFailTDAO().getURL('/')
        except DataFailureException as ex:
            self.assertEqual(ex.url, '/')
            self.assertEqual(ex.status, 0)

    def test_ssl_non_valid_ignore(self):
        response = SSLBadIgnoreTDAO().getURL('/ok', {})
        self.assertEqual(response.status, 200)
        self.assertEqual(response.data, b'Ok')



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from unittest import TestCase
from restclients_core import models
from datetime import datetime
import gc


class TestModelBase(TestCase):
    def test_override_init(self):
        class ModelTest(models.Model):
            bools = models.BooleanField()

            def __init__(self, *args, **kwargs):
                pass

        model = ModelTest()
        model.bools = True
        self.assertTrue(model.bools)

    def test_field_types(self):
        class ModelTest(models.Model):
            bools = models.BooleanField()
            chars = models.CharField(max_length=20)
            dates = models.DateField()
            datetimes = models.DateTimeField()
            decimals = models.DecimalField()
            floats = models.FloatField()
            ints = models.IntegerField()
            nullbools = models.NullBooleanField()
            posints = models.PositiveIntegerField()
            possmalls = models.PositiveSmallIntegerField()
            slugs = models.SlugField()
            smallints = models.SmallIntegerField()
            texts = models.TextField()
            texts2 = models.TextField()
            times = models.TimeField()
            urls = models.URLField()

        model = ModelTest()

        now = datetime.now()

        model.bools = True
        model.chars = "Char value"
        model.dates = now.date()
        model.datetimes = now
        model.decimals = 12.1
        model.floats = 1.2312
        model.ints = 21
        model.nullbools = False
        model.posints = 113234234
        model.possmalls = 10
        model.slugs = "ok"
        model.smallints = -1
        model.texts = "text string"
        model.texts2 = "making sure fields are different"
        model.times = now.time()
        model.urls = "http://example.com/path"

        self.assertEqual(model.bools, True)
        self.assertEqual(model.chars, "Char value")
        self.assertEqual(model.dates, now.date())
        self.assertEqual(model.datetimes, now)
        self.assertEqual(model.decimals, 12.1)
        self.assertEqual(model.floats, 1.2312)
        self.assertEqual(model.ints, 21)
        self.assertEqual(model.nullbools, False)
        self.assertEqual(model.posints, 113234234)
        self.assertEqual(model.possmalls, 10)
        self.assertEqual(model.slugs, "ok")
        self.assertEqual(model.smallints, -1)
        self.assertEqual(model.texts, "text string")
        self.assertEqual(model.texts2, "making sure fields are different")
        self.assertEqual(model.times, now.time())
        self.assertEqual(model.urls, "http://example.com/path")

        del (model.urls)
        self.assertIsNone(model.urls)

    def test_2_fields_2_instances(self):
        class ModelTest(models.Model):
            f1 = models.TextField()
            f2 = models.TextField()

        m1 = ModelTest()
        m2 = ModelTest()

        m1.f1 = "m1_f1"
        m1.f2 = "m1_f2"
        m2.f1 = "m2_f1"
        m2.f2 = "m2_f2"

        self.assertEqual(m1.f1, "m1_f1")
        self.assertEqual(m1.f2, "m1_f2")
        self.assertEqual(m2.f1, "m2_f1")
        self.assertEqual(m2.f2, "m2_f2")

    def test_init_fields(self):
        class ModelTest(models.Model):
            f1 = models.TextField()
            f2 = models.BooleanField()

        m1 = ModelTest(f1="Input value", f2=True)

        self.assertEqual(m1.f1, "Input value")
        self.assertEqual(m1.f2, True)

    def test_default_values(self):
        class ModelTest(models.Model):
            f1 = models.TextField(default="Has Default")
            f2 = models.TextField()

        m1 = ModelTest()
        m2 = ModelTest(f1="override")

        self.assertEqual(m1.f1, "Has Default")
        self.assertEqual(m1.f2, None)

        self.assertEqual(m2.f1, "override")
        self.assertEqual(m2.f2, None)

    def test_char_choices(self):
        CHOICES = (('ok', 'OK!'), ('not_ok', 'Not OK!'))

        class ModelTest(models.Model):
            f1 = models.CharField(default='ok', choices=CHOICES)
            f2 = models.CharField(default='ok2')

        m1 = ModelTest()
        self.assertEqual(m1.f1, 'ok')
        self.assertEqual(m1.get_f1_display(), 'OK!')

        m1.f1 = 'not_ok'
        self.assertEqual(m1.f1, 'not_ok')
        self.assertEqual(m1.get_f1_display(), 'Not OK!')

        with self.assertRaises(AttributeError):
            m1.get_f2_display()

    def test_memory_leak(self):
        class MemTest2(models.Model):
            pass

        class MemTest(models.Model):
            mt0 = models.ForeignKey(MemTest2)

        def build():
            m = MemTest()
            m.mt0 = MemTest2()
            m.mt0.x = m

        def match_count():
            count = 0
            for obj in gc.get_objects():
                obj_str = repr(obj)
                if obj_str.find('restclients_core.models.fields') >= 0:
                    count += 1
            return count

        build()
        gc.collect()
        starting = match_count()

        build()
        gc.collect()
        after_2 = match_count()

        self.assertEqual(starting, after_2)



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from unittest import TestCase
from commonconf import settings, override_settings
from restclients_core.thread import Thread
from mock import patch


class MockSQLite():
    vendor = 'sqlite'


class MockMySQL():
    vendor = 'mysql'


class TestThread(TestCase):
    @patch('restclients_core.thread.db_connection', MockMySQL)
    def test_use_threading_mysql(self):
        thread = Thread()
        self.assertTrue(thread._use_thread)

        with override_settings(RESTCLIENTS_DISABLE_THREADING=True):
            thread = Thread()
            self.assertFalse(thread._use_thread)

        with override_settings(RESTCLIENTS_DISABLE_THREADING=False):
            thread = Thread()
            self.assertTrue(thread._use_thread)

    @patch('restclients_core.thread.db_connection', MockSQLite)
    def test_use_threading_sqlite(self):
        thread = Thread()
        self.assertFalse(thread._use_thread)

        with override_settings(RESTCLIENTS_DISABLE_THREADING=True):
            thread = Thread()
            self.assertFalse(thread._use_thread)

        with override_settings(RESTCLIENTS_DISABLE_THREADING=False):
            thread = Thread()
            self.assertFalse(thread._use_thread)

    def test_use_threading_nodb(self):
        thread = Thread()
        self.assertFalse(thread._use_thread)

        with override_settings(RESTCLIENTS_USE_THREADING=True):
            thread = Thread()
            self.assertTrue(thread._use_thread)

        with override_settings(RESTCLIENTS_USE_THREADING=False):
            thread = Thread()
            self.assertFalse(thread._use_thread)



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

import random
import datetime
from restclients_core.util.mock import load_resource_from_path
from restclients_core.util.local_cache import (
    set_cache_value, get_cache_value)
from restclients_core.models import MockHTTP, CacheHTTP
from restclients_core.exceptions import (
    ImproperlyConfigured, DataFailureException)
from restclients_core.cache import NoCache
from restclients_core.util.performance import PerformanceDegradation
from importlib import import_module
from commonconf import settings
from urllib3 import connection_from_url
from urllib3.util import Timeout
from urllib3.util.retry import Retry
from urllib3.exceptions import HTTPError
from prometheus_client import Histogram, Counter
from logging import getLogger
from dateutil.parser import parse
from urllib.parse import urlparse
import time
import ssl

logger = getLogger(__name__)

# prepare for prometheus observations
prometheus_duration = Histogram('restclient_request_duration_seconds',
                                'Restclient request duration (seconds)',
                                ['service'])
prometheus_status = Histogram('restclient_response_status_code',
                              'Restclient web service response status code',
                              ['service'],
                              buckets=[100, 200, 300, 400, 500])
prometheus_timeout = Counter('restclient_request_timeout',
                             'Restclient web service request timeout count',
                             ['service'])
prometheus_ssl_error = Counter('restclient_request_ssl_error',
                               'Restclient web service SSL error count',
                               ['service'])


class DAO(object):
    """
    Base class for per-service interfaces.
    """
    _cache_instance = None

    def __init__(self):
        # format is ISO 8601
        log_start_str = self.get_service_setting("TIMING_START", None)
        log_end_str = self.get_service_setting("TIMING_END", None)

        if log_start_str is not None and log_end_str is not None:
            self.log_start = parse(log_start_str)
            self.log_end = parse(log_end_str)
        else:
            self.log_start = None
            self.log_end = None

        self.log_timing = self.get_service_setting("TIMING_LOG_ENABLED", False)
        self.logging_rate = float(self.get_service_setting("TIMING_LOG_RATE",
                                                           1.0))

    def service_name(self):
        """
        This method must be overridden to define your service's short name.

        This name is used in multiple places.  The Mock DAO uses it in path
        names for file, and the Django app for browsing services uses it as
        part of the URL.
        """
        raise Exception("service_name must be defined per DAO")

    def _custom_headers(self, method, url, headers, body):
        """
        This method can be overridden to add headers to a request.  For
        example, a Bearer header can be added if a service uses OAuth tokens.
        """
        # to handle things like adding a bearer token
        pass

    def _custom_response_edit(self, method, url, headers, body, response):
        """
        This method allows a service to edit a response.

        If you want to do this, you probably really want to use
        _edit_mock_response - this method will operate on Live resources.
        """
        if self.get_implementation().is_mock():
            delay = self.get_setting("MOCKDATA_DELAY", 0.0)
            time.sleep(delay)
            self._edit_mock_response(method, url, headers, body, response)

    def _edit_mock_response(self, method, url, headers, body, response):
        """
        Override this method to edit responses in mock resources.  This can be
        used to ensure datetime fields have useful values relative to now,
        or to provide more dynamic behavior for PUT/POST/DELETE requests.

        This method should edit the response object directly.  No return value.
        """
        pass

    def get_default_service_setting(self, key):
        """
        A hook for setting useful defaults.  For example, if you have a host
        your service almost always uses, you can have this method return that
        value when passed 'HOST'.
        """
        return None

    def getURL(self, url, headers={}):
        """
        Request a URL using the HTTP method GET
        """
        return self._load_resource("GET", url, headers, None)

    def postURL(self, url, headers={}, body=None):
        """
        Request a URL using the HTTP method POST.
        """
        return self._load_resource("POST", url, headers, body)

    def putURL(self, url, headers, body=None):
        """
        Request a URL using the HTTP method PUT.
        """
        return self._load_resource("PUT", url, headers, body)

    def patchURL(self, url, headers, body):
        """
        Request a URL using the HTTP method PATCH.
        """
        return self._load_resource("PATCH", url, headers, body)

    def deleteURL(self, url, headers=None):
        """
        Request a URL using the HTTP method DELETE.
        """
        return self._load_resource("DELETE", url, headers, None)

    def service_mock_paths(self):
        """
        If your web service client ships with mock resources, override this
        method to return a list of top level paths where they can be found.

        e.g. If your resource is in
        /users/my/my_client/resources/client/file/hello.json
        you should generate ["/users/my/my_client/resources"]
        """
        return []

    def _load_resource(self, method, url, headers, body):
        start_time = time.time()
        service = self.service_name()

        bad_response = PerformanceDegradation.get_response(service, url)
        if bad_response:
            return bad_response

        custom_headers = self._custom_headers(method, url, headers, body)
        if custom_headers:
            headers.update(custom_headers)

        is_cacheable = self._is_cacheable(method, url, headers, body)

        cache = self.get_cache()

        if is_cacheable:
            cache_response = cache.getCache(service, url, headers)
            if cache_response:
                if "response" in cache_response:
                    self._log(service=service, url=url, method=method,
                              response=cache_response["response"],
                              cached=True, start_time=start_time)
                    return cache_response["response"]
                if "headers" in cache_response:
                    headers = cache_response["headers"]

        backend = self.get_implementation()

        response = backend.load(method, url, headers, body)

        self.prometheus_duration(time.time() - start_time)
        self.prometheus_status(response)

        self._custom_response_edit(method, url, headers, body, response)

        if is_cacheable:
            cache_post_response = cache.processResponse(service, url, response)
            if cache_post_response is not None:
                if "response" in cache_post_response:
                    self._log(service=service, url=url, method=method,
                              response=response, cached=True,
                              start_time=start_time)
                    return cache_post_response["response"]

        self._log(service=service, url=url, method=method, response=response,
                  cached=False, start_time=start_time)

        return response

    def prometheus_duration(self, duration):
        """
        Override this method if you have service-specific logic
        around response times
        """
        self.prometheus_duration_observation(duration)

    def prometheus_status(self, response):
        """
        Override this method to insert service-specific logic
        before setting the response status code observation

        e.g., If the service applies special meaning to, say, 404 response
        that might not make sense to observe
        """
        self.prometheus_status_observation(response.status)

    def prometheus_duration_observation(self, duration):
        prometheus_duration.labels(self.service_name()).observe(duration)

    def prometheus_status_observation(self, status):
        # status category buckets
        prometheus_status.labels(self.service_name()).observe(
            (int(status) // 100) * 100)

    def get_cache(self):
        if DAO._cache_instance is None:
            implementation = self.get_setting("DAO_CACHE_CLASS", None)
            DAO._cache_instance = self._getModule(implementation, NoCache)
        return DAO._cache_instance

    def clear_cached_response(self, url):
        self.get_cache().deleteCache(self.service_name(), url)

    def get_implementation(self):
        implementation = self.get_service_setting("DAO_CLASS", None)

        # Handle the easy built-ins
        if "Live" == implementation:
            return self._get_live_implementation()

        if "Mock" == implementation:
            return self._get_mock_implementation()

        # Legacy settings support
        live = "restclients.dao_implementation.{}.Live".format(
            self.service_name())
        mock = "restclients.dao_implementation.{}.File".format(
            self.service_name())

        if live == implementation:
            return self._get_live_implementation()

        if mock == implementation:
            return self._get_mock_implementation()

        if implementation:
            return self._getModule(implementation, None,
                                   [self.service_name(), self])

        return self._get_mock_implementation()

    def _is_cacheable(self, method, url, headers, body=None):
        if method == "GET":
            return True
        return False

    def _ok_status_codes():
        return [200, 201, 202, 204]

    def _error_status_codes():
        return []

    def _get_live_implementation(self):
        return LiveDAO(self.service_name(), self)

    def _get_mock_implementation(self):
        return MockDAO(self.service_name(), self)

    def get_service_setting(self, key, default=None):
        if default is None:
            default = self.get_default_service_setting(key)

        service_key = "{}_{}".format(self.service_name().upper(), key)

        if hasattr(settings, "RESTCLIENTS_{}".format(service_key)):
            return self.get_setting(service_key, default)
        else:
            return self.get_setting(key, default)

    def get_setting(self, key, default=None):
        key = "RESTCLIENTS_{}".format(key)
        return getattr(settings, key, default)

    def _getModule(self, value, default_class, args=[]):
        if not value:
            return default_class()

        module, attr = value.rsplit('.', 1)
        try:
            mod = import_module(module)
        except ImportError as e:
            raise ImproperlyConfigured(
                "Error importing module {}: {}".format(module, e))
        try:
            config_module = getattr(mod, attr)
        except AttributeError:
            raise ImproperlyConfigured(
                "Module {} missing {} class".format(module, attr))
        return config_module(*args)

    def _log(self, *args, **kwargs):
        if not self.should_log():
            return

        from_cache = 'yes' if kwargs.get('cached') else 'no'
        cache_class = self.get_cache().__class__.__qualname__ if (
            kwargs.get('cached')) else 'None'
        response = kwargs.get('response')
        total_time = time.time() - kwargs.get('start_time')
        msg = (("service:{} method:{} url:{} status:{} from_cache:{} " +
                "cache_class:{} time:{}").format(
                   kwargs.get('service'), kwargs.get('method'),
                   kwargs.get('url'), response.status,
                   from_cache, cache_class, total_time))
        logger.info(msg)

    def should_log(self):

        if self.log_start is not None and self.log_end is not None:
            if not self.log_start < datetime.datetime.now() < self.log_end:
                return False

        if not self.log_timing:
            return False

        if random.random() >= self.logging_rate:
            return False

        return True


class DAOImplementation(object):
    def __init__(self, service_name, dao):
        self._service_name = service_name
        self.dao = dao

    def is_live(self):
        return False

    def is_mock(self):
        return False


class LiveDAO(DAOImplementation):
    """
    Loads response objects by fetching resources from an HTTP(s) server.
    """
    pools = {}

    def is_live(self):
        return True

    def load(self, method, url, headers, body):
        pool = self.get_pool()
        timeout = pool.timeout
        try:
            return pool.urlopen(
                method, url, body=body, headers=headers,
                timeout=timeout,
                pool_timeout=timeout.connect_timeout)
            # will block for 1 sec if no connection is available
            # then raise EmptyPoolError
        except ssl.SSLError as err:
            self._prometheus_ssl_error()
            raise
        except HTTPError as err:
            status = 0
            self._prometheus_timeout()
            raise DataFailureException(url, status, err)

    def get_pool(self):
        service = self.dao.service_name()
        if service not in LiveDAO.pools:
            pool = self.create_pool()
            LiveDAO.pools[service] = pool

        return LiveDAO.pools[service]

    def create_pool(self):
        """
        Return a ConnectionPool instance of given host
        """
        ca_certs = self.dao.get_setting("CA_BUNDLE",
                                        "/etc/ssl/certs/ca-bundle.crt")
        cert_file = self.dao.get_service_setting("CERT_FILE", None)
        host = self.dao.get_service_setting("HOST")
        key_file = self.dao.get_service_setting("KEY_FILE", None)
        verify_https = self.dao.get_service_setting("VERIFY_HTTPS")
        ssl_context = self.dao.get_service_setting("SSL_CONTEXT")

        if verify_https is None:
            verify_https = True

        timeout = Timeout(connect=self._get_connect_timeout(),
                          read=self._get_timeout())
        kwargs = {
            "retries": Retry(total=1, connect=0, read=0, redirect=1),
            "timeout": timeout,
            "maxsize": self._get_max_pool_size(),
            "block": True,
        }

        if key_file is not None and cert_file is not None:
            kwargs["key_file"] = key_file
            kwargs["cert_file"] = cert_file

        if urlparse(host).scheme == "https":
            kwargs["ssl_version"] = self.dao.get_service_setting(
                "SSL_VERSION", ssl.PROTOCOL_TLS)

            if verify_https:
                kwargs["cert_reqs"] = "CERT_REQUIRED"
                kwargs["ca_certs"] = ca_certs
            else:
                kwargs["cert_reqs"] = "CERT_NONE"

            if ssl_context is not None:
                kwargs["ssl_context"] = ssl_context

        return connection_from_url(host, **kwargs)

    def _get_connect_timeout(self):
        """
        The maximum amount of time (in seconds) to wait for a connection
        attempt to a server to succeed
        """
        return float(self.dao.get_service_setting("CONNECT_TIMEOUT",
                     self.dao.get_setting("DEFAULT_CONNECT_TIMEOUT", 3)))

    def _get_timeout(self):
        """
        The maximum amount of time (in seconds) to wait between
        consecutive READ operations for a response from the server.
        """
        return float(self.dao.get_service_setting("TIMEOUT",
                     self.dao.get_setting("DEFAULT_TIMEOUT", 10)))

    def _get_max_pool_size(self):
        """
        The maximum connections per host.
        """
        return int(self.dao.get_service_setting("POOL_SIZE",
                   self.dao.get_setting("DEFAULT_POOL_SIZE", 10)))

    def _prometheus_timeout(self):
        prometheus_timeout.labels(self.dao.service_name()).inc()

    def _prometheus_ssl_error(self):
        prometheus_ssl_error.labels(self.dao.service_name()).inc()


class MockDAO(DAOImplementation):
    """
    Loads response objects based on file content.
    """
    paths = []

    def is_mock(self):
        return True

    @classmethod
    def register_mock_path(cls, path):
        if path not in MockDAO.paths:
            MockDAO.paths.append(path)

    def get_registered_paths(self):
        return MockDAO.paths

    def _get_mock_paths(self):
        return self.get_registered_paths() + self.dao.service_mock_paths()

    def load(self, method, url, headers, body):
        service = self._service_name

        cache_key = "{}-{}".format(service, url)
        value = get_cache_value(cache_key)
        if value:
            return value

        for path in self._get_mock_paths():
            response = load_resource_from_path(
                path, service, "file", url, headers)

            if response and response.status != 404:
                set_cache_value(cache_key, response)
                return response

        response = MockHTTP()
        response.status = 404
        response.reason = "Not Found"

        set_cache_value(cache_key, response)
        return response



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from restclients_core.tests.dao_implementation.test_backend import TDAO
from unittest import TestCase, skipUnless
from commonconf import override_settings
import datetime


class TestTimingLog(TestCase):

    @skipUnless(hasattr(TestCase, 'assertLogs'), 'Python < 3.4')
    @override_settings(RESTCLIENTS_TIMING_START=(datetime.datetime.now() +
                       datetime.timedelta(1)).isoformat(),
                       RESTCLIENTS_TIMING_END=(datetime.datetime.now() +
                       datetime.timedelta(2)).isoformat(),
                       RESTCLIENTS_TIMING_LOG_ENABLED=True,
                       RESTCLIENTS_TIMING_LOG_RATE=1.0)
    def test_dont_log_date_range(self):
        with self.assertRaises(AssertionError):
            with self.assertLogs('restclients_core.dao', level='INFO') as cm:
                response = TDAO().getURL('/ok')
                self.assertEqual(len(cm.output), 0)

    @skipUnless(hasattr(TestCase, 'assertLogs'), 'Python < 3.4')
    @override_settings(RESTCLIENTS_TIMING_START=(datetime.datetime.now() -
                       datetime.timedelta(1)).isoformat(),
                       RESTCLIENTS_TIMING_END=(datetime.datetime.now() +
                       datetime.timedelta(2)).isoformat(),
                       RESTCLIENTS_TIMING_LOG_ENABLED=True,
                       RESTCLIENTS_TIMING_LOG_RATE=1.0)
    def test_log_date_range(self):
        with self.assertLogs('restclients_core.dao', level='INFO') as cm:
            response = TDAO().getURL('/ok')
            self.assertEqual(len(cm.output), 1)

    @skipUnless(hasattr(TestCase, 'assertLogs'), 'Python < 3.4')
    @override_settings(RESTCLIENTS_TIMING_LOG_ENABLED=False,
                       RESTCLIENTS_TIMING_LOG_RATE=1.0,
                       RESTCLIENTS_BACKEND_TEST_TIMING_LOG_ENABLED=True)
    def test_per_service_config(self):
        with self.assertLogs('restclients_core.dao', level='INFO') as cm:
            response = TDAO().getURL('/ok')
            self.assertEqual(len(cm.output), 1)



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

import datetime
import time


class BaseField(object):
    default = None

    def __init__(self, *args, **kwargs):
        self.values = {}
        self.dynamics = set()

        if "default" in kwargs:
            self.default = kwargs["default"]

        super(BaseField, self).__init__()

    def __get__(self, instance, owner):
        key = self._key_for_instance(instance)

        field_key = self._key_for_instance(self)

        try:
            value = instance._get_value(field_key)
            return value
        except KeyError:
            return self.default

    def __set__(self, instance, value):
        field_key = self._key_for_instance(self)
        instance._set_value(field_key, value)
        instance._track_field(self)

    def __delete__(self, instance):
        field_key = self._key_for_instance(self)
        instance._delete(field_key)

    def _key_for_instance(self, instance):
        if not hasattr(instance, "__rcm_timestamp"):
            setattr(instance, "__rcm_timestamp", time.time())

        key = "{}-{}".format(
            id(instance), getattr(instance, "__rcm_timestamp"))
        return key

    def clean(self, instance):
        pass


class CharField(BaseField):
    has_choices = False

    def __init__(self, *args, **kwargs):
        nullable = False
        if "null" in kwargs and kwargs["null"]:
            nullable = True

        if not nullable:
            self.default = u""

        if "choices" in kwargs:
            self.has_choices = True
            self.choices = kwargs["choices"]

        super(CharField, self).__init__(*args, **kwargs)

    def get_display(self, instance):
        val = self.__get__(instance, None)

        if not self.has_choices:
            raise Exception("No choices on field")

        for opt in self.choices:
            if opt[0] == val:
                return opt[1]


class BooleanField(BaseField):
    pass


class DateField(BaseField):
    def clean(self, instance):
        value = self.__get__(instance, None)

        if isinstance(value, type(datetime.datetime.now())):
            self.__set__(instance, value.date())


class DateTimeField(BaseField):
    pass


class DecimalField(BaseField):
    pass


class FloatField(BaseField):
    pass


class IntegerField(BaseField):
    pass


class NullBooleanField(BaseField):
    pass


class PositiveIntegerField(BaseField):
    pass


class PositiveSmallIntegerField(BaseField):
    pass


class SlugField(BaseField):
    pass


class SmallIntegerField(BaseField):
    pass


class TextField(BaseField):
    pass


class TimeField(BaseField):
    pass


class URLField(BaseField):
    pass


class ForeignKey(BaseField):
    pass



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from restclients_core.exceptions import DataFailureException
import math
import time


def retry(ExceptionToCheck, tries=4, delay=3, backoff=2, status_codes=[],
          logger=None):
    """
    Decorator function for retrying the decorated function,
    using an exponential or fixed backoff.

    Original: https://wiki.python.org/moin/PythonDecoratorLibrary#Retry

    ExceptionToCheck: the exception to check. Can be a tuple of
        exceptions to check
    tries: number of times to try (not retry) before giving up
    delay: initial delay between tries in seconds
    backoff: backoff multiplier
    status_codes: list of http status codes to check for retrying, only applies
         when ExceptionToCheck is a DataFailureException
    logger: logging.Logger instance
    """
    if backoff is None or backoff <= 0:
        raise ValueError("backoff must be a number greater than 0")

    tries = math.floor(tries)
    if tries < 0:
        raise ValueError("tries must be a number 0 or greater")

    if delay is None or delay <= 0:
        raise ValueError("delay must be a number greater than 0")

    def deco_retry(f):
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)

                except ExceptionToCheck as err:
                    if (type(err) is DataFailureException and
                            len(status_codes) and
                            err.status not in status_codes):
                        raise

                    if logger:
                        logger.warning(
                            "{}: {}, Retrying in {} seconds.".format(
                                f.__name__, err, mdelay))

                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff

            return f(*args, **kwargs)

        return f_retry

    return deco_retry



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from commonconf import settings
import threading

try:
    from django.db import connection as db_connection
except ImportError:
    db_connection = None


"""
This is a wrapper around threading.Thread, but it will only actually thread
if django configuration is enabled.  Otherwise, it will be an object with the
same api where start() just calls run().
"""


class Thread(threading.Thread):
    _use_thread = False
    parent = None

    def __init__(self, *args, **kwargs):
        self.parent = threading.current_thread()

        if self._db_compatible:
            self._use_thread = not getattr(
                settings, "RESTCLIENTS_DISABLE_THREADING", False)

        else:
            self._use_thread = getattr(
                settings, "RESTCLIENTS_USE_THREADING", False)

        super().__init__(*args, **kwargs)

    @property
    def _db_compatible(self):
        return db_connection is not None and db_connection.vendor != 'sqlite'

    def start(self):
        if self._use_thread:
            super().start()

        else:
            # Needed to test failures in the threads.
            # But it can't be on all the time - sqlite dbs aren't shared to
            # threads.
            if getattr(settings, "RESTCLIENTS_USE_INLINE_THREADS", False):
                super().start()
                super().join()
            else:
                self.run()

    """
    Subclasses should call final() in their overridden run().
    """
    def run(self):
        # Override to perform specific tasks.
        self.final()

    def join(self):
        if self._use_thread:
            return super().join()
        return True

    def final(self):
        if (self._use_thread and db_connection is not None and
                not db_connection.in_atomic_block):
            db_connection.close()


class GenericPrefetchThread(Thread):
    method = None

    def run(self):
        if self.method is not None:
            try:
                self.method()
            except Exception as ex:
                # Errors in prefetching should also manifest during actual
                # processing, where they can be handled appropriately
                pass

        self.final()


def generic_prefetch(method, args):
    def ret():
        return method(*args)
    return ret



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

import re
import weakref
from restclients_core.models.fields import (BooleanField, CharField, DateField,
                                            DateTimeField, DecimalField,
                                            FloatField, ForeignKey,
                                            IntegerField, NullBooleanField,
                                            PositiveIntegerField,
                                            PositiveSmallIntegerField,
                                            SlugField, SmallIntegerField,
                                            TextField, TimeField, URLField,
                                            BaseField)


class MockHTTP(object):
    """
    An alternate object to HTTPResponse, for non-HTTP DAO
    implementations to use.  Implements the API of HTTPResponse
    as needed.
    """
    status = 0
    data = ""
    headers = {}

    def read(self):
        """
        Returns the document body of the request.
        """
        return self.data

    def getheader(self, field, default=''):
        """
        Returns the HTTP response header field, case insensitively
        """
        if self.headers:
            for header in self.headers:
                if field.lower() == header.lower():
                    return self.headers[header]

        return default


class CacheHTTP(MockHTTP):
    """
    A class which implements the interface of HTTPResponse but signals to
    clients that the object they are receiving is from a cache.
    """

    cache_class = None

    def get_cache_class(self):
        return self.cache_class


class Model(object):
    initialized = False

    def __init__(self, *args, **kwargs):
        self._init()

        super(Model, self).__init__()

        for key in kwargs:
            setattr(self, key, kwargs[key])

    def _init(self):
        # Some models overrode __init__ without calling super.  This keeps
        # those models from failing now
        if not self.initialized:
            self.initialized = True
            self._dynamic_fields = set()
            self._field_values = {}

    def _set_value(self, field_id, value):
        self._init()
        self._field_values[field_id] = value

    def _get_value(self, field_id):
        self._init()
        return self._field_values[field_id]

    def _delete(self, field_id):
        self._init()
        del (self._field_values[field_id])

    def _track_field(self, field):
        self._dynamic_fields.add(weakref.ref(field))

    def __getattribute__(self, name):
        # This is in place to catch get_<attribute>_display.  If there's
        # alrady a defined attribute here, just return it.  Otherwise if the
        # attribute name matches our pattern, and we can find that name in our
        # Class's dictionary, we try to let that field type get the right
        # value.  then we return a function that returns the value.
        try:
            base_value = super(Model, self).__getattribute__(name)
            return base_value
        except AttributeError as ex:
            base_value = None
            original_exception = ex

        match = re.match('get_(.*)_display', name)
        if match:
            try:
                field = self.__class__.__dict__[match.group(1)]

                value = field.get_display(self)
                return lambda: value

            except Exception:
                pass

        raise original_exception

    def clean_fields(self):
        for ref in self._dynamic_fields:
            field = ref()
            if field:
                field.clean(self)

        pass

    def __str__(self):
        return ", ".join([
            "{}: {}".format(k, getattr(self, k))
            for k, v in sorted(self.__class__.__dict__.items())
            if issubclass(type(v), BaseField)])


PROTECT = None



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from unittest import TestCase
from threading import current_thread
from restclients_core.util.local_cache import (
    local_cache, set_cache_value, get_cache_value, LOCAL_CACHE)


class TestCache(TestCase):
    def test_context(self):
        thread = current_thread()
        with local_cache() as lc:
            v1 = test_method1("init")

            self.assertTrue(thread in LOCAL_CACHE)
            v2 = test_method1("nope, old cached value")
            self.assertEqual(v2, "init")

        self.assertFalse(thread in LOCAL_CACHE)

    def test_decorator(self):
        thread = current_thread()

        @local_cache()
        def use_the_cache():
            v1 = test_method1("init decorator")

            self.assertTrue(thread in LOCAL_CACHE)
            v2 = test_method1("nope, old cached value")
            self.assertEqual(v2, "init decorator")

        use_the_cache()
        self.assertFalse(thread in LOCAL_CACHE)

    def test_no_cache(self):
        v1 = test_method1("init none")
        self.assertEqual(v1, "init none")

        v2 = test_method1("second, none")
        self.assertEqual(v2, "second, none")


def test_method1(val):
    key = "test_method1_key"
    value = get_cache_value(key)
    if value:
        return value

    set_cache_value(key, val)
    return val



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from threading import current_thread

LOCAL_CACHE = {}


class local_cache(object):
    """
    A decorator or context manager that will cache some values for the
    lifetime of the decorator/context manager.
    """
    def _function(self, func, *args, **kwargs):
        def wrapped(*args, **kwargs):
            try:
                self.__enter__()
                value = func(*args, **kwargs)
            finally:
                self.__exit__()
            return value
        return wrapped

    # This handles function decorators + class decorators.  Needs to switch
    # on argument type :(
    def __call__(self, decorated, *args, **kwargs):
        if isinstance(decorated, type):
            raise Exception("Can only decorate a function, not a class")

        return self._function(decorated, *args, **kwargs)

    # These handle context managers, and are used by the others
    def __enter__(self, *args, **kwargs):
        thread = current_thread()

        LOCAL_CACHE[thread] = {}
        return self

    def __exit__(*args, **kwargs):
        thread = current_thread()
        del LOCAL_CACHE[thread]


def _get_local_cache():
    thread = current_thread()
    if thread in LOCAL_CACHE:
        return LOCAL_CACHE[thread]
    if hasattr(thread, 'parent'):
        thread = thread.parent
        if thread in LOCAL_CACHE:
            return LOCAL_CACHE[thread]

    return


def set_cache_value(key, value):
    """
    If local cache is being used, this will set a cache value for the
    lifetime of that cache.
    """
    cache = _get_local_cache()
    if cache is None:
        return

    cache[key] = value


def get_cache_value(key):
    """
    If local cache is being used, this will get the cache value, if it exists.
    """
    cache = _get_local_cache()
    if cache is None:
        return

    if key in cache:
        return cache[key]



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from restclients_core.models import CacheHTTP
from unittest import TestCase


class TestCacheHTTP(TestCase):

    def test_cache_http(self):
        obj = CacheHTTP()

        obj.cache_class = self.__class__
        self.assertEqual(obj.get_cache_class(), self.__class__)


class SubClassTest(TestCacheHTTP):

    def test_cache_http(self):
        obj = CacheHTTP()

        obj.cache_class = self.__class__
        self.assertEqual(obj.get_cache_class(), SubClassTest)



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from threading import current_thread
from restclients_core.models import MockHTTP
import time


class PerformanceDegradation(object):
    _problem_data = {}
    problems = None

    @classmethod
    def set_problems(obj, problems):
        thread = current_thread()
        PerformanceDegradation._problem_data[thread] = problems

    @classmethod
    def clear_problems(obj):
        thread = current_thread()
        PerformanceDegradation._problem_data[thread] = None

    @classmethod
    def get_problems(obj):
        thread = current_thread()

        if thread in PerformanceDegradation._problem_data:
            return PerformanceDegradation._problem_data[thread]

        if hasattr(thread, 'parent'):
            thread = thread.parent
            if thread in PerformanceDegradation._problem_data:
                return PerformanceDegradation._problem_data[thread]

        return None

    @classmethod
    def get_response(obj, service, url):
        problems = PerformanceDegradation.get_problems()
        if not problems:
            return

        delay = problems.get_load_time(service)
        if delay:
            time.sleep(float(delay))

        status = problems.get_status(service)
        content = problems.get_content(service)

        if content and not status:
            status = 200

        if status:
            response = MockHTTP()
            response.status = int(status)

            if content:
                response.data = content

            return response

        return None



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from unittest import TestCase
from restclients_core.util.mock import convert_to_platform_safe


class TestPlatformSafe(TestCase):
    def test_convert_to_platform_safe(self):
        name = "CreateEditor?Name=M%20M&Email=x@uw.edu&Password="
        self.assertEqual(convert_to_platform_safe(name),
                         "CreateEditor_Name_M%20M_Email_x_uw.edu_Password_")

        name = "r/2013,spring,T%20BUS,310,A,1,1.json"
        self.assertEqual(convert_to_platform_safe(name),
                         "r/2013_spring_T%20BUS_310_A_1_1.json")

        name = "sections?per_page=50&include=students@"
        self.assertEqual(convert_to_platform_safe(name),
                         "sections_per_page_50_include_students_")

        name = "admins?page=2&per_page=10"
        self.assertEqual(convert_to_platform_safe(name),
                         "admins_page_2_per_page_10")

        name = "enrollments?role=student"
        self.assertEqual(convert_to_platform_safe(name),
                         "enrollments_role_student")

        name = "2012,autumn,MATH,120/"
        self.assertEqual(convert_to_platform_safe(name),
                         "2012_autumn_MATH_120/")

        name = "sws/file/student/v4/course/2013,spring,PHIL,600/A"
        self.assertEqual(convert_to_platform_safe(name),
                         "sws/file/student/v4/course/2013_spring_PHIL_600/A")



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from unittest import TestCase
from restclients_core.exceptions import DataFailureException
from restclients_core.util.retry import retry


class RetryableError(Exception):
    pass


class AnotherRetryableError(Exception):
    pass


class UnexpectedError(Exception):
    pass


class RetryTest(TestCase):

    def test_bad_args(self):
        self.assertRaises(TypeError, retry, RetryableError, tries=None)
        self.assertRaises(ValueError, retry, RetryableError, delay=None)
        self.assertRaises(ValueError, retry, RetryableError, backoff=None)
        self.assertRaises(ValueError, retry, RetryableError, tries=-1)
        self.assertRaises(ValueError, retry, RetryableError, delay=0)
        self.assertRaises(ValueError, retry, RetryableError, backoff=0)

    def test_no_retry(self):
        self.counter = 0

        @retry(RetryableError, tries=4, delay=0.1)
        def succeeds():
            self.counter += 1
            return 'success'

        r = succeeds()

        self.assertEqual(r, 'success')
        self.assertEqual(self.counter, 1)

    def test_retry_once(self):
        self.counter = 0

        @retry(RetryableError, tries=4, delay=0.1)
        def fails_once():
            self.counter += 1
            if self.counter < 2:
                raise RetryableError('failed')
            else:
                return 'success'

        r = fails_once()
        self.assertEqual(r, 'success')
        self.assertEqual(self.counter, 2)

    def test_limit_reached(self):
        self.counter = 0

        @retry(RetryableError, tries=4, delay=0.1)
        def always_fails():
            self.counter += 1
            raise RetryableError('failed')

        self.assertRaises(RetryableError, always_fails)
        self.assertEqual(self.counter, 4)

    def test_multiple_exception_types(self):
        self.counter = 0

        @retry((RetryableError, AnotherRetryableError), tries=4, delay=0.1)
        def raise_multiple_exceptions():
            self.counter += 1
            if self.counter == 1:
                raise RetryableError('a retryable error')
            elif self.counter == 2:
                raise AnotherRetryableError('another retryable error')
            else:
                return 'success'

        r = raise_multiple_exceptions()
        self.assertEqual(r, 'success')
        self.assertEqual(self.counter, 3)

    def test_unexpected_exception_does_not_retry(self):
        self.counter = 0

        @retry(RetryableError, tries=4, delay=0.1)
        def raise_unexpected_error():
            self.counter += 1
            raise UnexpectedError('unexpected error')

        self.assertRaises(UnexpectedError, raise_unexpected_error)
        self.assertEqual(self.counter, 1)

    def test_dfe_with_no_status(self):
        self.counter = 0

        @retry(DataFailureException, tries=4, delay=0.1)
        def not_found():
            self.counter += 1
            raise DataFailureException('/', 404, '')

        self.assertRaises(DataFailureException, not_found)
        self.assertEqual(self.counter, 4)

    def test_no_retry_with_status(self):
        self.counter = 0

        @retry(DataFailureException, status_codes=[500], tries=4, delay=0.1)
        def not_found():
            self.counter += 1
            raise DataFailureException('/', 404, '')

        self.assertRaises(DataFailureException, not_found)
        self.assertEqual(self.counter, 1)

    def test_retry_once_with_status(self):
        self.counter = 0

        @retry(DataFailureException, status_codes=[500], tries=4, delay=0.1)
        def fails_once():
            self.counter += 1
            if self.counter < 2:
                raise DataFailureException('/', 500, '')
            else:
                return 'success'

        r = fails_once()
        self.assertEqual(r, 'success')
        self.assertEqual(self.counter, 2)

    def test_limit_reached_with_status(self):
        self.counter = 0

        @retry(DataFailureException, status_codes=[500], tries=4, delay=0.1)
        def always_fails():
            self.counter += 1
            raise DataFailureException('/', 500, '')

        self.assertRaises(DataFailureException, always_fails)
        self.assertEqual(self.counter, 4)

    def test_multiple_status_codes(self):
        self.counter = 0

        @retry(DataFailureException, status_codes=[500, 503, 504],
               tries=4, delay=0.1)
        def raise_multiple_exceptions():
            self.counter += 1
            if self.counter == 1:
                raise DataFailureException('/', 500, '')
            elif self.counter == 2:
                raise DataFailureException('/', 503, '')
            else:
                return 'success'

        r = raise_multiple_exceptions()
        self.assertEqual(r, 'success')
        self.assertEqual(self.counter, 3)

    def test_multiple_exceptions_and_multiple_status_codes(self):
        self.counter = 0

        @retry((DataFailureException, RetryableError),
               status_codes=[500, 503, 504], tries=4, delay=0.1)
        def raise_multiple_exceptions():
            self.counter += 1
            if self.counter == 1:
                raise DataFailureException('/', 500, '')
            elif self.counter == 2:
                raise DataFailureException('/', 503, '')
            elif self.counter == 3:
                raise RetryableError()
            else:
                return 'success'

        r = raise_multiple_exceptions()
        self.assertEqual(r, 'success')
        self.assertEqual(self.counter, 4)



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from unittest import TestCase
from restclients_core import models


class TestModelString(TestCase):
    def test_model_string(self):
        class ModelTest(models.Model):
            number = models.CharField(max_length=8)
            one = models.IntegerField()

        m = ModelTest(number="number", one=1)
        self.assertEqual("{}".format(m), "number: number, one: 1")



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from unittest import TestCase, skipIf
from restclients_core.dao import DAO
from os.path import abspath, dirname
from restclients_core.dao import MockDAO
from restclients_core.exceptions import DataFailureException


class TDAO(DAO):
    def service_name(self):
        return 'testing'

    def service_mock_paths(self):
        # tests/dao_implementation/resources/
        return [abspath(dirname(__file__) + "/resources/")]


class TestMock(TestCase):
    def test_found_resource(self):
        response = TDAO().getURL('/found.json', {})
        self.assertEqual(response.status, 200)
        self.assertEqual(response.read(), b'{"OK": true }\n')

    def test_missing_resource(self):
        response = TDAO().getURL('/missing.json', {})
        self.assertEqual(response.status, 404)

    def test_file_headers(self):
        response = TDAO().getURL('/with_headers.json', {})
        self.assertEqual(response.status, 202)

        self.assertEqual(response.headers["Custom"], "My Custom Value")
        self.assertEqual(response.headers.get("Custom"),
                         "My Custom Value")

    def test_plain_file_headers(self):
        response = TDAO().getURL('/with_only_headers.json', {})
        self.assertEqual(response.status, 200)

        self.assertEqual(response.headers["Custom2"], "My Custom Value 2")

    def test_registered_paths(self):
        response = TDAO().getURL('/override.json', {})
        self.assertEqual(response.status, 200)
        self.assertEqual(response.read(), b'{"override": false }\n')

        override = abspath(dirname(__file__) + "/resource_override/")
        MockDAO.register_mock_path(override)

        response = TDAO().getURL('/override.json', {})
        self.assertEqual(response.status, 200)
        self.assertEqual(response.read(), b'{"override": true }\n')

    def test_binary_data(self):
        response = TDAO().getURL('/image.jpg', {})
        self.assertEqual(response.status, 200)

    def test_params_none(self):
        response = TDAO().getURL('/search?'
                                 'first=a&second=b.POST')
        self.assertEqual(response.status, 200)

    def test_out_of_order_params(self):
        response = TDAO().getURL('/search?'
                                 'first=a&second=b&third=c&fourth=d')
        self.assertEqual(response.status, 200)

    def test_extra_params(self):
        response = TDAO().getURL('/search?'
                                 'first=a&second=b&third=c&fourth=d&fifth=e')
        self.assertEqual(response.status, 404)

    def test_quoted_params(self):
        response = TDAO().getURL('/search?'
                                 'first=a&second=a%3Ab%3Ac')
        self.assertEqual(response.status, 200)

    def test_no_body_file(self):
        response = TDAO().getURL('/search?'
                                 'first=abc')
        self.assertEqual(response.status, 200)

    def test_multiple_files(self):
        with self.assertRaises(DataFailureException):
            TDAO().getURL('/search?'
                          'first=a&second=b')

    def test_quote_in_file_path(self):
        response = TDAO().getURL('/test%3folder/test.json')
        self.assertEqual(response.status, 200)

    def test_quote_in_file_path_and_file(self):
        response = TDAO().getURL('/test%3folder/test.json')
        self.assertEqual(response.status, 200)

    @skipIf(True, "I'm not sure that this case actually exists - ezturner")
    def test_quote_in_file_path_and_url_but_not_file(self):
        response = TDAO().getURL('/test%3folder/test%3man.json')
        self.assertEqual(response.status, 200)



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from restclients_core.tests.dao_implementation.test_backend import TDAO
from unittest import TestCase
from prometheus_client import generate_latest, REGISTRY


class TestPrometheusObservations(TestCase):
    def test_prometheus_observation(self):
        response = TDAO().getURL('/ok')

        metrics = generate_latest(REGISTRY).decode('utf-8')
        self.assertRegex(
            metrics,
            r'.*\nrestclient_request_duration_seconds_bucket{le="0.005",'
            r'service="backend_test"} [1-9].*', metrics)
        self.assertRegex(
            metrics,
            r'.*\nrestclient_response_status_code_bucket{le="200.0",'
            r'service="backend_test"} [1-9].*', metrics)
        self.assertIn('restclient_request_timeout_total counter', metrics)
        self.assertIn('restclient_request_ssl_error_total counter', metrics)



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from unittest import TestCase, skipUnless
from commonconf import override_settings
from restclients_core.dao import DAO, MockDAO
from restclients_core.cache import NoCache
from restclients_core.models import MockHTTP, CacheHTTP
from restclients_core.exceptions import ImproperlyConfigured


class TDAO(DAO):
    def service_name(self):
        return "backend_test"

    def get_default_service_setting(self, key):
        if "DAO_CLASS" == key:
            return ('restclients_core.tests.dao_implementation.'
                    'test_backend.Backend')


class E1DAO(TDAO):
    def get_default_service_setting(self, key):
        if "DAO_CLASS" == key:
            return ('restclients_core.tests.dao_implementation.'
                    'test_backendX.Backend')


class E2DAO(TDAO):
    def get_default_service_setting(self, key):
        if "DAO_CLASS" == key:
            return ('restclients_core.tests.dao_implementation.'
                    'test_backend.BackendX')


class TCache():
    def getCache(self, service, url, headers):
        if url == '/ok':
            response = CacheHTTP()
            response.cache_class = self.__class__
            response.status = 200
            response.data = 'ok - GET'
            return {'response': response}

    def processResponse(self, service, url, response):
        response.status = 404
        response.cache_class = self.__class__
        return {'response': response}


class TestBackend(TestCase):
    def setUp(self):
        DAO._cache_instance = None

    def test_get(self):
        response = TDAO().getURL('/ok')
        self.assertEqual(response.data, 'ok - GET')

    def test_post(self):
        response = TDAO().postURL('/ok')
        self.assertEqual(response.data, 'ok - POST')

    def test_put(self):
        response = TDAO().putURL('/ok', {}, '')
        self.assertEqual(response.data, 'ok - PUT')

    def test_delete(self):
        response = TDAO().deleteURL('/ok')
        self.assertEqual(response.data, 'ok - DELETE')

    def test_patch(self):
        response = TDAO().patchURL('/ok', {}, '')
        self.assertEqual(response.data, 'ok - PATCH')

    def test_error_level1(self):
        self.assertRaises(ImproperlyConfigured, E1DAO().getURL, '/ok')

    def test_error_level2(self):
        self.assertRaises(ImproperlyConfigured, E2DAO().getURL, '/ok')

    @override_settings(RESTCLIENTS_BACKEND_TEST_FOO=True,
                       RESTCLIENTS_FOO=False,
                       RESTCLIENTS_BAR=True)
    def test_service_settings(self):
        dao = TDAO()
        self.assertEqual(dao.get_setting('FOO'), False)
        self.assertEqual(dao.get_setting('BAR'), True)

        self.assertEqual(dao.get_service_setting('FOO'), True)
        self.assertEqual(dao.get_service_setting('BAR'), True)

    @skipUnless(hasattr(TestCase, 'assertLogs'), 'Python < 3.4')
    @override_settings(RESTCLIENTS_TIMING_LOG_ENABLED=True,
                       RESTCLIENTS_TIMING_LOG_RATE=1.0)
    def test_log(self):
        with self.assertLogs('restclients_core.dao', level='INFO') as cm:
            response = TDAO().getURL('/ok')
            self.assertEqual(len(cm.output), 1)
            (msg, time) = cm.output[0].split(' time:')
            self.assertEqual(msg,
                             'INFO:restclients_core.dao:service:backend_test '
                             'method:GET url:/ok status:200 from_cache:no'
                             ' cache_class:None')
            self.assertGreater(float(time), 0)

        with self.assertLogs('restclients_core.dao', level='INFO') as cm:
            response = TDAO().putURL('/api', {}, '')
            self.assertEqual(len(cm.output), 1)
            (msg, time) = cm.output[0].split(' time:')
            self.assertEqual(msg,
                             'INFO:restclients_core.dao:service:backend_test '
                             'method:PUT url:/api status:200 from_cache:no'
                             ' cache_class:None')
            self.assertGreater(float(time), 0)

    @skipUnless(hasattr(TestCase, 'assertLogs'), 'Python < 3.4')
    @override_settings(RESTCLIENTS_TIMING_LOG_ENABLED=True,
                       RESTCLIENTS_TIMING_LOG_RATE=1.0,
                       RESTCLIENTS_DAO_CACHE_CLASS=(
                           'restclients_core.tests.dao_implementation.'
                           'test_backend.TCache'))
    def test_log_cache(self):
        # Cached response
        with self.assertLogs('restclients_core.dao', level='INFO') as cm:
            response = TDAO().getURL('/ok')
            self.assertEqual(len(cm.output), 1)
            (msg, time) = cm.output[0].split(' time:')
            self.assertEqual(msg,
                             'INFO:restclients_core.dao:service:backend_test '
                             'method:GET url:/ok status:200 from_cache:yes'
                             ' cache_class:TCache')
            self.assertGreater(float(time), 0)

        # Cached post response
        with self.assertLogs('restclients_core.dao', level='INFO') as cm:
            response = TDAO().getURL('/ok2')
            self.assertEqual(len(cm.output), 1)
            (msg, time) = cm.output[0].split(' time:')
            self.assertEqual(msg,
                             'INFO:restclients_core.dao:service:backend_test '
                             'method:GET url:/ok2 status:404 from_cache:yes'
                             ' cache_class:TCache')
            self.assertGreater(float(time), 0)

    @override_settings(RESTCLIENTS_DAO_CACHE_CLASS=(
                           'restclients_core.tests.dao_implementation.'
                           'test_backend.TCache'))
    def test_cache_backend(self):
        cache1 = TDAO().get_cache()
        cache2 = TDAO().get_cache()
        self.assertTrue(cache1 == cache2, 'Cache objects are same instance')

        cache3 = E1DAO().get_cache()
        self.assertTrue(cache1 == cache3, 'Cache objects are same instance')

        cache4 = E2DAO().get_cache()
        self.assertTrue(cache1 == cache4, 'Cache objects are same instance')


class Backend(MockDAO):
    def load(self, method, url, headers, body):
        response = MockHTTP()
        response.status = 200
        response.data = "ok - {}".format(method)
        return response



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from unittest import TestCase
from restclients_core import models
from datetime import datetime


class TestModelCleanMethods(TestCase):
    def test_date_clean(self):
        class ModelTest(models.Model):
            d1 = models.DateField()

        now = datetime.now()
        m1 = ModelTest(d1=now)

        self.assertEqual(now, m1.d1)
        m1.clean_fields()
        self.assertEqual(now.date(), m1.d1)



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

import re
import os
from os.path import isfile, join, dirname
import json
from urllib.parse import unquote
from restclients_core.exceptions import DataFailureException
from restclients_core.models import MockHTTP


def load_resource_from_path(resource_dir,
                            service_name,
                            implementation_name,
                            url,
                            headers):
    if url == "///":
        # Just a placeholder to put everything else in an else.
        # If there are things that need dynamic work, they'd go here
        pass
    else:
        response = MockHTTP()
        response.status = 404
        response.headers = None
        RESOURCE_ROOT = os.path.join(resource_dir,
                                     service_name,
                                     implementation_name)
        orig_file_path = RESOURCE_ROOT + url

        handle = open_file(orig_file_path)
        if handle is not None:
            response.status = 200
            response.data = handle.read()
            handle.close()

        header_handle = open_file(orig_file_path + ".http-headers")
        if header_handle is not None:
            __read_header(header_handle, response, service_name)
            header_handle.close()

        # Check query permutations even on success
        # so that if there are multiple files we throw an exception
        if "?" in url:
            handle = attempt_open_query_permutations(
                url, orig_file_path, False)
            if handle is not None:
                if response.status == 404:
                    response.status = 200
                    response.data = handle.read()
                handle.close()

            header_handle = attempt_open_query_permutations(
                url, orig_file_path, True)
            if header_handle is not None:
                if response.headers is None:
                    __read_header(header_handle, response, service_name)
                header_handle.close()

        return response


def __read_header(header_handle, response, service_name):
    try:
        data = header_handle.read()
        file_values = json.loads(data)

        if response.headers is None:
            response.headers = {
                "X-Data-Source": service_name + " file mock data",
            }

        if "headers" in file_values:
            response.headers.update(file_values['headers'])

            if 'status' in file_values:
                response.status = file_values['status']
        else:
            response.headers.update(file_values)
    except UnicodeDecodeError:
        pass
    except IOError:
        pass


def convert_to_platform_safe(dir_file_name):
    """
    :param dir_file_name: a string to be processed
    :return: a string with all the reserved characters replaced
    """
    return re.sub(r'[\?|<>=:*,;+&"@$]', '_', dir_file_name)


def open_file(orig_file_path):
    """
    Taking in a file path, attempt to open mock data files with it.
    """
    unquoted = unquote(orig_file_path)
    paths = [
        convert_to_platform_safe(orig_file_path),
        "{}/index.html".format(convert_to_platform_safe(orig_file_path)),
        orig_file_path,
        "{}/index.html".format(orig_file_path),
        convert_to_platform_safe(unquoted),
        "{}/index.html".format(convert_to_platform_safe(unquoted)),
        unquoted,
        "{}/index.html".format(unquoted),
        ]

    file_path = None
    handle = None
    for path in paths:
        try:
            file_path = path
            handle = open(path, "rb")
            break
        except IOError:
            pass

    return handle


def attempt_open_query_permutations(url, orig_file_path, is_header_file):
    """
    Attempt to open a given mock data file with different permutations of the
    query parameters
    """
    directory = dirname(convert_to_platform_safe(orig_file_path)) + "/"

    # get all filenames in directory
    try:
        filenames = [f for f in os.listdir(directory)
                     if isfile(join(directory, f))]
    except OSError:
        return

    # ensure that there are not extra parameters on any files
    if is_header_file:
        filenames = [f for f in filenames if ".http-headers" in f]
        filenames = [f for f in filenames if
                     _compare_file_name(orig_file_path + ".http-headers",
                                        directory,
                                        f)]
    else:
        filenames = [f for f in filenames if ".http-headers" not in f]
        filenames = [f for f in filenames if _compare_file_name(orig_file_path,
                                                                directory,
                                                                f)]

    url_parts = url.split("/")
    url_parts = url_parts[len(url_parts) - 1].split("?")

    base = url_parts[0]
    params = url_parts[1]

    params = params.split("&")

    # check to ensure that the base url matches
    filenames = [f for f in filenames if f.startswith(base)]

    params = [convert_to_platform_safe(unquote(p)) for p in params]

    # ensure that all parameters are there
    for param in params:
        filenames = [f for f in filenames if param in f]

    # if we only have one file, return it
    if len(filenames) == 1:
        path = join(directory, filenames[0])
        return open_file(path)

    # if there is more than one file, raise an exception
    if len(filenames) > 1:
        raise DataFailureException(url,
                                   "Multiple mock data files matched the " +
                                   "parameters provided!",
                                   404)


def _compare_file_name(orig_file_path, directory, filename):
    return (len(unquote(orig_file_path)) - len(unquote(directory)) ==
            len(unquote(filename)))



# Copyright 2025 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

import os
from setuptools import setup

README = """
See the README on `GitHub
<https://github.com/uw-it-aca/uw-restclients-core>`_.
"""

version_path = 'restclients_core/VERSION'
VERSION = open(os.path.join(os.path.dirname(__file__), version_path)).read()
VERSION = VERSION.replace("\n", "")

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

url = "https://github.com/uw-it-aca/uw-restclients-core"
setup(
    name='UW-RestClients-Core',
    version=VERSION,
    packages=['restclients_core'],
    author="UWIT Student & Educational Technology Services",
    author_email="aca-it@uw.edu",
    include_package_data=True,
    install_requires=['commonconf',
                      'urllib3',
                      'python-dateutil',
                      'prometheus-client',
                      'mock'],
    license='Apache License, Version 2.0',
    description=('Core code for clients of a variety of RESTful web services '
                 'at the University of Washington'),
    long_description=README,
    url=url,
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
)



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


class NoCache(object):
    """
    A cache implementation that never caches.
    """
    def getCache(self, service, url, headers):
        return None

    def processResponse(self, service, url, response):
        pass

    def deleteCache(self, service, url):
        return None



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from commonconf import override_settings


def use_mock(*args):
    kw_args = {}

    for service in args:
        key = "RESTCLIENTS_{}_DAO_CLASS".format(
            service.service_name().upper())

        kw_args[key] = 'Mock'

    return override_settings(**kw_args)



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from unittest import TestCase
from commonconf import settings
from restclients_core.dao import DAO
from restclients_core.util.decorators import use_mock


class TDAO1(DAO):
    def service_name(self):
        return 'td1'


class TDAO2(DAO):
    def service_name(self):
        return 'td2'


class TestMockOverride(TestCase):
    def test_multi_service(self):
        self.assertEqual(getattr(settings, "RESTCLIENTS_TD1_DAO_CLASS", None),
                         None)

        with use_mock(TDAO1(), TDAO2()):
            self.assertEqual(settings.RESTCLIENTS_TD1_DAO_CLASS, 'Mock')
            self.assertEqual(settings.RESTCLIENTS_TD2_DAO_CLASS, 'Mock')

        self.assertEqual(getattr(settings, "RESTCLIENTS_TD1_DAO_CLASS", None),
                         None)



# Copyright 2026 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


class ImproperlyConfigured(Exception):
    """Required setting is missing"""
    pass


class DataFailureException(Exception):
    """
    This exception means there was an error fetching content
    in one of the rest clients.  You can get the url that failed
    with .url, the status of the error with .status, and any
    message with .msg
    """
    def __init__(self, url, status, msg):
        self.url = url
        self.status = status
        self.msg = msg

    def __str__(self):
        return (
            "Error fetching {}.  Status code: {}.  Message: {}.".format(
                self.url, self.status, self.msg))


class InvalidRegID(Exception):
    """Exception for invalid regid."""
    pass


class InvalidNetID(Exception):
    """Exception for invalid netid."""
    pass


class InvalidEmployeeID(Exception):
    """Exception for invalid employee id."""
    pass
